from collections.abc import Sequence
from typing import Literal

from draive import DataModel, Field


class SchemaNestedModel(DataModel):
    value: Literal["A", "B", "C"] = Field(description="selection")
    other_value: None


class SchemaModel(DataModel):
    str_value: str
    int_value: int
    float_value: float
    bool_value: bool
    list_value: Sequence[str]
    optional_value: str | None = Field(description="alternative")
    nested: SchemaNestedModel = Field(aliased="nested_value", description="alternative")


json_schema: str = """\
{
  "type": "object",
  "properties": {
    "str_value": {
      "type": "string"
    },
    "int_value": {
      "type": "integer"
    },
    "float_value": {
      "type": "number"
    },
    "bool_value": {
      "type": "boolean"
    },
    "list_value": {
      "type": "array",
      "items": {
        "type": "string"
      }
    },
    "optional_value": {
      "oneOf": [
        {
          "type": "string"
        },
        {
          "type": "null"
        }
      ],
      "description": "alternative"
    },
    "nested_value": {
      "type": "object",
      "properties": {
        "value": {
          "type": "string",
          "enum": [
            "A",
            "B",
            "C"
          ],
          "description": "selection"
        },
        "other_value": {
          "type": "null"
        }
      },
      "required": [
        "value",
        "other_value"
      ],
      "description": "alternative"
    }
  },
  "required": [
    "str_value",
    "int_value",
    "float_value",
    "bool_value",
    "list_value",
    "optional_value",
    "nested_value"
  ]
}\
"""


def test_json_schema() -> None:
    assert json_schema == SchemaModel.json_schema(indent=2)


simplified_schema: str = """\
{
  "str_value": "string",
  "int_value": "integer",
  "float_value": "number",
  "bool_value": "boolean",
  "list_value": [
    "string"
  ],
  "optional_value": "string|null(alternative)",
  "nested_value": {
    "value": "'A'|'B'|'C'(selection)",
    "other_value": "null"
  }
}\
"""


def test_simplified_schema() -> None:
    assert simplified_schema == SchemaModel.simplified_schema(indent=2)



from typing import Any

from haiway import State, ctx

from draive.guardrails.privacy.types import (
    GuardrailsAnonymizedContent,
    GuardrailsContentAnonymizing,
)
from draive.multimodal import Multimodal, MultimodalContent

__all__ = ("GuardrailsAnonymization",)


class GuardrailsAnonymization(State):
    @classmethod
    async def anonymize(
        cls,
        content: Multimodal,
        /,
        **extra: Any,
    ) -> GuardrailsAnonymizedContent:
        return await ctx.state(cls).anonymizing(MultimodalContent.of(content))

    anonymizing: GuardrailsContentAnonymizing



from asyncio import AbstractEventLoop, CancelledError, Task, TimerHandle, gather, get_running_loop
from collections.abc import Sequence
from inspect import isfunction
from typing import Protocol, cast, final, overload, runtime_checkable
from uuid import UUID, uuid4

from haiway import MISSING, AsyncQueue, Missing, State, ctx, not_missing

from draive.agents.definition import AgentMessage, AgentNode
from draive.agents.errors import AgentException
from draive.agents.idle import IdleMonitor
from draive.agents.node import Agent, AgentOutput
from draive.agents.runner import AgentRunner
from draive.helpers import VolatileMemory
from draive.multimodal import (
    Multimodal,
    MultimodalContent,
)
from draive.parameters import DataModel
from draive.utils import Memory

__all__ = (
    "AgentWorkflow",
    "AgentWorkflowIdle",
    "AgentWorkflowInput",
    "AgentWorkflowInvocation",
    "AgentWorkflowOutput",
    "workflow",
)


type AgentWorkflowInput = AgentMessage | AgentException
type AgentWorkflowOutput[Result: DataModel | State | str] = (
    Sequence[AgentMessage] | AgentMessage | Result | None
)


@runtime_checkable
class AgentWorkflowStateInitializer[AgentWorkflowState](Protocol):
    def __call__(self) -> AgentWorkflowState: ...


@runtime_checkable
class AgentWorkflowInvocation[AgentWorkflowState, AgentWorkflowResult: DataModel | State | str](
    Protocol
):
    async def __call__(
        self,
        memory: Memory[AgentWorkflowState, AgentWorkflowState],
        input: AgentWorkflowInput,  # noqa: A002
    ) -> AgentWorkflowOutput[AgentWorkflowResult]: ...


class AgentWorkflowIdle(AgentException):
    pass


@final
class AgentWorkflow[AgentWorkflowState, AgentWorkflowResult: DataModel | State | str]:
    def __init__(
        self,
        node: AgentNode,
        invocation: AgentWorkflowInvocation[
            AgentWorkflowState,
            AgentWorkflowResult,
        ],
        state_initializer: AgentWorkflowStateInitializer[AgentWorkflowState],
    ) -> None:
        self.node: AgentNode = node
        self._invocation: AgentWorkflowInvocation[
            AgentWorkflowState,
            AgentWorkflowResult,
        ] = invocation
        self._state_initializer: AgentWorkflowStateInitializer[AgentWorkflowState] = (
            state_initializer
        )

    async def __call__(
        self,
        memory: Memory[AgentWorkflowState, AgentWorkflowState],
        input: AgentWorkflowInput,  # noqa: A002
    ) -> AgentWorkflowOutput[AgentWorkflowResult]:
        return await self._invocation(
            memory=memory,
            input=input,
        )

    async def run(
        self,
        input: Multimodal,  # noqa: A002
        state: AgentWorkflowState | None = None,
        timeout: float = 120,  # default timeout is 2 minutes
    ) -> AgentWorkflowResult:
        return await WorkflowRunner[AgentWorkflowState, AgentWorkflowResult].run(
            self,
            input=input,
            memory=VolatileMemory(initial=state or self._state_initializer()),
            timeout=timeout,
        )

    def address(
        self,
        content: Multimodal,
        /,
        *_content: Multimodal,
        addressee: AgentNode | None = None,
    ) -> AgentMessage:
        return AgentMessage(
            identifier=uuid4(),
            sender=MISSING,
            recipient=self.node,
            addressee=addressee,
            content=MultimodalContent.of(content, *_content),
            responding=None,
        )


class PartialAgentWorkflowWrapper[AgentWorkflowState](Protocol):
    def __call__[AgentWorkflowResult: DataModel | State | str](
        self,
        invocation: AgentWorkflowInvocation[AgentWorkflowState, AgentWorkflowResult],
    ) -> AgentWorkflow[AgentWorkflowState, AgentWorkflowResult]: ...


@overload
def workflow[AgentWorkflowState](
    node: AgentNode,
    /,
    *,
    state: AgentWorkflowStateInitializer[AgentWorkflowState],
) -> PartialAgentWorkflowWrapper[AgentWorkflowState]: ...


@overload
def workflow[AgentWorkflowState](
    *,
    name: str | None = None,
    description: str,
    state: AgentWorkflowStateInitializer[AgentWorkflowState],
) -> PartialAgentWorkflowWrapper[AgentWorkflowState]: ...


def workflow[AgentWorkflowState](
    node: AgentNode | None = None,
    *,
    name: str | None = None,
    description: str | None = None,
    state: AgentWorkflowStateInitializer[AgentWorkflowState],
) -> PartialAgentWorkflowWrapper[AgentWorkflowState]:
    assert node is None or (  # nosec: B101
        name is None and description is None
    ), "Can't specify both agent node and name/description"
    assert (  # nosec: B101
        description is not None or node is not None
    ), "Either agent node or description has to be provided"

    def wrap[AgentWorkflowResult: DataModel | State | str](
        invocation: AgentWorkflowInvocation[AgentWorkflowState, AgentWorkflowResult],
    ) -> AgentWorkflow[AgentWorkflowState, AgentWorkflowResult]:
        assert isfunction(invocation), "workflow has to be defined from a function"  # nosec: B101

        workflow_node: AgentNode = node or AgentNode(
            name=name or invocation.__qualname__,
            description=description or "",
        )

        def initialize_agent() -> Agent:
            agent_memory: Memory[AgentWorkflowState, AgentWorkflowState] = VolatileMemory(
                initial=state()
            )

            async def agent(message: AgentMessage) -> AgentOutput:
                match await cast(  # pyright fails to properly type this function
                    AgentWorkflowInvocation[AgentWorkflowState, AgentWorkflowResult],
                    invocation,
                )(
                    memory=agent_memory,
                    input=message,
                ):
                    case None:
                        pass

                    case tuple() as messages:
                        return messages

                    case AgentMessage() as message:
                        return message

                    case result:
                        return message.respond(
                            MultimodalContent.of(
                                result if isinstance(result, DataModel) else str(result)
                            )
                        )

            return agent

        # workflow can run as an agent with limited capabilities
        workflow_node.associate(
            initialize_agent,
            concurrent=False,  # workflow can't be concurrent
        )

        return AgentWorkflow[AgentWorkflowState, AgentWorkflowResult](
            node=workflow_node,
            invocation=invocation,
            state_initializer=state,
        )

    return wrap


@runtime_checkable
class WorkflowRunnerOutput(Protocol):
    def __call__(
        self,
        messages: AgentMessage,
        /,
        *_messages: AgentMessage,
    ) -> None: ...


@runtime_checkable
class WorkflowRunnerResultOutput[WorkflowResult: DataModel | State | str](Protocol):
    def __call__(
        self,
        result: WorkflowResult,
        /,
    ) -> None: ...


class WorkflowRunner[WorkflowState, WorkflowResult: DataModel | State | str]:
    @classmethod
    async def run(
        cls,
        workflow: AgentWorkflow[WorkflowState, WorkflowResult],
        /,
        input: Multimodal,  # noqa: A002
        memory: Memory[WorkflowState, WorkflowState],
        timeout: float,
    ) -> WorkflowResult:
        return await cls(
            workflow=workflow,
            memory=memory,
            loop=get_running_loop(),
        ).execute(
            input=input,
            timeout=timeout,
        )

    def __init__(
        self,
        workflow: AgentWorkflow[WorkflowState, WorkflowResult],
        memory: Memory[WorkflowState, WorkflowState],
        loop: AbstractEventLoop,
    ) -> None:
        self._loop: AbstractEventLoop = loop
        self._history: list[AgentMessage] = []
        self._workflow_queue: AsyncQueue[AgentMessage | AgentException] = AsyncQueue()
        self._workflow_memory: Memory[WorkflowState, WorkflowState] = memory
        self._workflow: AgentWorkflow[WorkflowState, WorkflowResult] = workflow
        self._agent_runners: dict[UUID, AgentRunner] = {}
        self._status: IdleMonitor = IdleMonitor()
        self._result: WorkflowResult | BaseException | Missing = MISSING

    def __del__(self) -> None:
        self.finish(result=CancelledError())

    @overload
    def send(
        self,
        error: AgentException,
        /,
    ) -> None: ...

    @overload
    def send(
        self,
        messages: AgentMessage,
        /,
        *_messages: AgentMessage,
    ) -> None: ...

    def send(
        self,
        messages: AgentMessage | AgentException,
        /,
        *_messages: AgentMessage,
    ) -> None:
        pending: list[AgentMessage | AgentException] = [messages, *_messages]
        for _ in pending:  # ensure not idle when received a message
            self._status.enter_task()
        self._history.extend([message for message in pending if isinstance(message, AgentMessage)])
        self._workflow_queue.enqueue(*pending)

    @property
    def finished(self) -> bool:
        return not_missing(self._result)

    def finish(
        self,
        result: WorkflowResult | BaseException,
    ) -> None:
        if not_missing(self._result):
            return  # already finished

        ctx.log_debug("Finishing workflow with result: %s", result)

        self._result = result

        for runner in self._agent_runners.values():
            runner.finish()

        self._workflow_queue.finish()

    async def _wait(self) -> None:
        await gather(
            *[runner.wait() for runner in self._agent_runners.values()],
            return_exceptions=False,
        )

    async def finalize(
        self,
        result: WorkflowResult | BaseException,
    ) -> None:
        self.finish(result=result)
        await self._wait()

    async def execute(  # noqa: PLR0912
        self,
        input: Multimodal,  # noqa: A002
        timeout: float,
    ) -> WorkflowResult:
        assert not self._agent_runners, "WorkflowRunner can run only once!"  # nosec: B101

        if self._workflow_queue.is_finished or not_missing(self._result):
            raise RuntimeError("WorkflowRunner can run only once!")

        async with ctx.scope(self._workflow.node.__str__()):

            def on_timeout() -> None:
                self.finish(result=TimeoutError())

            timeout_handle: TimerHandle = self._loop.call_later(
                delay=timeout,
                callback=on_timeout,
            )

            self.send(self._workflow.address(input))

            idle_monitor_task: Task[None] = ctx.spawn(self._idle_monitor)

            try:
                async for element in self._workflow_queue:
                    try:
                        match element:
                            case AgentMessage() as message:
                                if message.recipient.identifier == self._workflow.node.identifier:
                                    await self._handle(message)

                                elif runner := self._agent_runners.get(
                                    message.recipient.identifier
                                ):
                                    runner.send(message)

                                else:
                                    spawned_runner: AgentRunner = AgentRunner.run(
                                        message.recipient,
                                        output=self.send,
                                        status=self._status.nested(),
                                    )
                                    self._agent_runners[message.recipient.identifier] = (
                                        spawned_runner
                                    )
                                    spawned_runner.send(message)

                            case error:
                                await self._handle(error)

                    finally:
                        self._status.exit_task()

            except CancelledError:
                pass  # just end on cancel

            finally:  # cleanup status when exiting
                timeout_handle.cancel()  # cancel the timeout
                idle_monitor_task.cancel()  # finish idle monitor
                self._status.exit_all()

            await self._wait()  # wait for completion of all runners

            match self._result:
                case BaseException() as exc:
                    raise exc

                case Missing():
                    raise RuntimeError("Invalid workflow state")

                case result:
                    return result

    async def _handle(
        self,
        input: AgentWorkflowInput,  # noqa: A002
        /,
    ) -> None:
        try:
            match await self._workflow(
                memory=self._workflow_memory,
                input=input,
            ):
                case None:
                    pass  # nothing to do

                case [*messages]:
                    self.send(*[message._with_sender(self._workflow.node) for message in messages])  # pyright: ignore[reportPrivateUsage]

                case AgentMessage() as message:
                    self.send(message._with_sender(self._workflow.node))  # pyright: ignore[reportPrivateUsage]

                case result:
                    self.finish(result=cast(WorkflowResult, result))

        except BaseException as exc:
            self.finish(result=exc)

    async def _idle_monitor(self) -> None:
        while not self.finished:  # run until finished
            await self._status.wait_idle()  # wait until all agents become idle

            if self.finished:  # check if that is not due to finishing
                return  # workflow might be finished at this stage

            ctx.log_warning("Detected workflow idle state, attempting to resume...")

            # when workflow should still be running (did not provided result yet)
            # but no agent is working then notify workflow node about the error
            self.send(AgentWorkflowIdle())



from collections.abc import Sequence
from typing import cast

from draive.embedding import Embedded, ImageEmbedding, TextEmbedding
from draive.evaluation import EvaluationScore, EvaluationScoreValue, evaluator
from draive.multimodal import Multimodal, MultimodalContent, MultimodalTagElement
from draive.multimodal.media import MediaData
from draive.similarity.score import vector_similarity_score
from draive.stages import Stage

__all__ = (
    "image_vector_similarity_evaluator",
    "similarity_evaluator",
    "text_vector_similarity_evaluator",
)


INSTRUCTION: str = """\
You are evaluating the provided content according to the defined criteria.

<INSTRUCTION>
Compare the REFERENCE and the EVALUATED content by carefully examining them, then rate\
 the EVALUATED content using solely a similarity metric according to the EVALUATION_CRITERIA.
Think step by step and provide explanation of the score before the final score.
Use the explained RATING scale and the requested FORMAT to provide the result.
</INSTRUCTION>

<EVALUATION_CRITERIA>
Evaluated metric is similarity - the degree of semantic similarity between the REFERENCE\
 and the EVALUATED content.
</EVALUATION_CRITERIA>
{guidelines}
<RATING>
Assign a similarity score using exact name of one of the following values:
- "poor" is very low similarity, the content is completely unrelated in meaning.
- "good" is moderate similarity, the content share some common themes or ideas.
- "perfect" is very high similarity, the content is very close in meaning\
 or convey the same information.
Use the "none" value for content that cannot be rated at all.
</RATING>

<FORMAT>
The final result containing only the numerical score value HAVE to be put inside a `RESULT` \
xml tag within the result i.e. `<RESULT>good</RESULT>`.
</FORMAT>
"""


@evaluator(name="similarity")
async def similarity_evaluator(
    evaluated: Multimodal,
    /,
    *,
    reference: Multimodal,
    guidelines: str | None = None,
) -> EvaluationScore:
    if not evaluated:
        return EvaluationScore(
            value=0,
            comment="Input was empty!",
        )

    if not reference:
        return EvaluationScore(
            value=0,
            comment="Reference was empty!",
        )

    completion: MultimodalContent = await Stage.completion(
        MultimodalContent.of(
            "<REFERENCE>",
            reference,
            "</REFERENCE>\n<EVALUATED>",
            evaluated,
            "</EVALUATED>",
        ),
        instruction=INSTRUCTION.format(
            guidelines=f"\n<GUIDELINES>\n{guidelines}\n</GUIDELINES>\n"
            if guidelines is not None
            else ""
        ),
    ).execute()

    if result := MultimodalTagElement.parse_first(
        "RESULT",
        content=completion,
    ):
        return EvaluationScore.of(
            cast(EvaluationScoreValue, result.content.to_str()),
            comment=completion.to_str(),
        )

    else:
        raise ValueError("Invalid evaluator result:\n%s", completion)


@evaluator(name="text_vector_similarity")
async def text_vector_similarity_evaluator(
    evaluated: str,
    /,
    *,
    reference: str,
) -> float:
    embedding: Sequence[Embedded[str]] = await TextEmbedding.embed([reference, evaluated])

    return vector_similarity_score(
        value_vector=embedding[0].vector,
        reference_vector=embedding[1].vector,
    )


@evaluator(name="image_vector_similarity")
async def image_vector_similarity_evaluator(
    evaluated: MediaData | bytes,
    /,
    *,
    reference: MediaData | bytes,
) -> float:
    evaluated_data: bytes
    match evaluated:
        case MediaData() as media:
            evaluated_data = media.data

        case raw_data:
            evaluated_data = raw_data

    reference_data: bytes
    match reference:
        case MediaData() as media:
            reference_data = media.data

        case raw_data:
            reference_data = raw_data

    embedding: Sequence[Embedded[bytes]] = await ImageEmbedding.embed(
        [reference_data, evaluated_data]
    )

    return vector_similarity_score(
        value_vector=embedding[0].vector,
        reference_vector=embedding[1].vector,
    )



import json
from asyncio import Task, TaskGroup
from base64 import b64decode, b64encode
from collections.abc import AsyncIterator, Mapping, Sequence
from typing import Any, Literal

from haiway import MISSING, AsyncQueue, Missing, ctx, without_missing
from openai.resources.beta.realtime.realtime import (
    AsyncRealtimeConnection,
    AsyncRealtimeConnectionManager,
)
from openai.types.beta.realtime import (
    ConversationItemContentParam,
)
from openai.types.beta.realtime.realtime_server_event import RealtimeServerEvent
from openai.types.beta.realtime.session_update_event_param import Session, SessionTool

from draive.instructions import Instruction
from draive.lmm import (
    LMMContext,
    LMMSession,
    LMMSessionOutput,
    LMMSessionOutputSelection,
    LMMStreamInput,
    LMMToolSelection,
    LMMToolSpecification,
)
from draive.lmm.types import (
    LMMCompletion,
    LMMInput,
    LMMSessionEvent,
    LMMStreamChunk,
    LMMToolRequest,
    LMMToolRequests,
    LMMToolResponse,
    LMMToolResponses,
)
from draive.multimodal import MediaData, MediaReference
from draive.multimodal.content import MultimodalContent
from draive.multimodal.meta import MetaContent
from draive.multimodal.text import TextContent
from draive.openai.api import OpenAIAPI
from draive.openai.config import OpenAIRealtimeConfig
from draive.openai.types import OpenAIException

__all__ = ("OpenAILMMSession",)


class OpenAILMMSession(OpenAIAPI):
    def lmm_session(self) -> LMMSession:
        return LMMSession(preparing=self.lmm_session_prepare)

    async def lmm_session_prepare(
        self,
        *,
        instruction: Instruction | None = None,
        initial_context: LMMContext | None = None,
        input_stream: AsyncIterator[LMMStreamInput],
        output: LMMSessionOutputSelection = "auto",
        tools: Sequence[LMMToolSpecification] | None = None,
        tool_selection: LMMToolSelection = "auto",
        config: OpenAIRealtimeConfig | None = None,
        **extra: Any,
    ) -> AsyncIterator[LMMSessionOutput]:
        realtime_config: OpenAIRealtimeConfig = config or ctx.state(OpenAIRealtimeConfig)
        with ctx.scope("openai_realtime", realtime_config):
            ctx.record(
                attributes={
                    "lmm.provider": "openai",
                    "lmm.model": realtime_config.model,
                    "lmm.temperature": realtime_config.temperature,
                    "lmm.max_tokens": realtime_config.max_tokens,
                    "lmm.tools": [tool["name"] for tool in tools] if tools else [],
                    "lmm.tool_selection": f"{tool_selection}",
                    "lmm.output": f"{output}",
                    "lmm.context": [element.to_str() for element in initial_context or []],
                }
            )

            modalities: list[Literal["text", "audio"]]
            match output:
                case "auto":
                    modalities = ["text", "audio"]

                case "text":
                    modalities = ["text"]

                case "audio":
                    modalities = ["text", "audio"]  # openai does not allow to use only audio

                case ["text", "audio"] | ["audio", "text"]:
                    modalities = ["text", "audio"]

                case _:
                    raise ValueError(f"Unsupported output: {output}")

            tool_choice: str
            match tool_selection:
                case "auto":
                    tool_choice = "auto"

                case "required":
                    tool_choice = "required"

                case "none":
                    tool_choice = "none"

                case tool:
                    tool_choice = tool["name"]

            session_tools: list[SessionTool] | Missing
            if tools is not None:
                session_tools = [
                    without_missing(
                        {
                            "type": "function",
                            "name": tool["name"],
                            "description": tool.get("description", MISSING),
                            "parameters": tool.get("parameters", MISSING),
                        },
                        typed=SessionTool,
                    )
                    for tool in tools
                ]

            else:
                session_tools = MISSING

            output_queue = AsyncQueue[LMMSessionOutput]()
            ctx.spawn(
                self._drive_connection,
                session_connection=self._client.beta.realtime.connect(
                    model=realtime_config.model,
                ),
                session_config=without_missing(
                    {
                        "instructions": Instruction.formatted(instruction)
                        if instruction is not None
                        else MISSING,
                        "temperature": realtime_config.temperature,
                        "modalities": modalities,
                        "input_audio_format": realtime_config.input_audio_format,
                        "output_audio_format": realtime_config.output_audio_format,
                        "turn_detection": {
                            "create_response": True,
                            "eagerness": realtime_config.vad_eagerness,
                            "interrupt_response": True,
                            "type": realtime_config.vad_type,
                        }
                        if realtime_config.vad_type is not MISSING
                        else MISSING,
                        "voice": realtime_config.voice,
                        "max_response_output_tokens": realtime_config.max_tokens,
                        "tools": session_tools,
                        "tool_choice": tool_choice if session_tools is not MISSING else MISSING,
                    },
                    typed=Session,
                ),
                initial_context=initial_context,
                input_stream=input_stream,
                output_stream=output_queue,
            )

            return output_queue

    @ctx.scope("connection", task_group=TaskGroup())
    async def _drive_connection(  # noqa: C901, PLR0912, PLR0915
        self,
        session_connection: AsyncRealtimeConnectionManager,
        *,
        session_config: Session,
        initial_context: LMMContext | None,
        input_stream: AsyncIterator[LMMStreamInput],
        output_stream: AsyncQueue[LMMSessionOutput],
    ):
        input_task: Task[None] | None = None
        try:
            async with session_connection as connection:  # connect
                try:  # prepare session
                    await connection.session.update(session=session_config)

                    if initial_context:  # initialize state
                        await self._send_context(
                            initial_context,
                            connection=connection,
                        )

                    del initial_context

                except BaseException as exc:
                    return output_stream.finish(exception=exc)

                audio_output_mime: str = (
                    f"audio/{session_config.get('output_audio_format', 'pcm16')}"
                )
                # start consuming inputs
                input_task = ctx.spawn(
                    self._handle_input,
                    connection=connection,
                    input_stream=input_stream,
                    output_stream=output_stream,
                )

                # start consuming outputs
                pending_tool_calls: set[str] = set()
                partial_tool_calls: dict[str, dict[str, str]] = {}
                while not output_stream.is_finished:
                    event: RealtimeServerEvent = await connection.recv()
                    match event.type:
                        case "response.audio.delta":
                            output_stream.enqueue(
                                LMMStreamChunk.of(
                                    MediaData.of(
                                        b64decode(event.delta),
                                        media=audio_output_mime,
                                    )
                                )
                            )

                        case "response.audio.done":
                            output_stream.enqueue(LMMSessionEvent.of("completed"))

                        case "response.function_call_arguments.delta":
                            partial_tool_calls[event.item_id]["arguments"] += event.delta

                        case "response.function_call_arguments.done":
                            tool_call: Mapping[str, str] = partial_tool_calls[event.item_id]
                            del partial_tool_calls[event.item_id]
                            pending_tool_calls.add(tool_call["call_id"])
                            output_stream.enqueue(
                                LMMToolRequest.of(
                                    tool_call["call_id"],
                                    tool=tool_call["tool"],
                                    arguments=json.loads(tool_call["arguments"])
                                    if tool_call["arguments"]
                                    else None,
                                )
                            )

                        case "response.text.delta":
                            output_stream.enqueue(LMMStreamChunk.of(event.delta))

                        case "response.text.done":
                            output_stream.enqueue(LMMSessionEvent.of("completed"))

                        case "response.output_item.added":
                            if event.item.type != "function_call":
                                continue  # skip other items

                            if event.item.id is None:
                                continue  # can't use events without item ids

                            if event.item.call_id is None:
                                continue  # skip tool calls without call id

                            if event.item.name is None:
                                continue  # skip tool calls without tool name

                            match event.item.status:
                                case "completed":
                                    pending_tool_calls.add(event.item.call_id)
                                    output_stream.enqueue(
                                        LMMToolRequest.of(
                                            event.item.call_id,
                                            tool=event.item.name,
                                            arguments=json.loads(event.item.arguments)
                                            if event.item.arguments
                                            else None,
                                        )
                                    )

                                case _:
                                    partial_tool_calls[event.item.id] = {
                                        "call_id": event.item.call_id,
                                        "tool": event.item.name,
                                        "arguments": event.item.arguments or "",
                                    }

                        case "conversation.item.created":
                            if event.item.type != "function_call_output":
                                continue  # skip other items

                            assert event.item.call_id is not None  # nosec: B101

                            if event.item.call_id not in pending_tool_calls:
                                continue  # skip tools which are not expected to respond

                            pending_tool_calls.remove(event.item.call_id)

                            if not pending_tool_calls:
                                # request a response after providing all tool results
                                await connection.response.create()

                        case "input_audio_buffer.speech_started":
                            # treat audio input start as an interrupt
                            output_stream.enqueue(LMMSessionEvent.of("interrupted"))
                            # skip pending tool results in that case?
                            pending_tool_calls.clear()

                        case "error":
                            raise OpenAIException(
                                f"Realtime processing error:"
                                f" {event.error.type} - {event.error.message}",
                                event,
                            )

                        case _:
                            pass  # skip other events

        except BaseException as exc:
            output_stream.finish(exc)

        else:
            output_stream.finish()

        finally:
            if input_task is not None:
                input_task.cancel()

    async def _send_input_chunk(
        self,
        chunk: LMMStreamChunk,
        /,
        *,
        connection: AsyncRealtimeConnection,
    ) -> None:
        # stream audio if we got only audio
        if chunk.content.is_media("audio"):
            for part in chunk.content.media("audio"):
                match part:
                    case MediaData() as media:
                        await connection.input_audio_buffer.append(
                            audio=b64encode(media.data).decode()
                        )

                    case _:
                        # skip not supported with a log to prevent connection break
                        ctx.log_error(
                            "OpenAI realtime input (MediaReference) not supported! Skipping..."
                        )

            if chunk.eod:
                await connection.input_audio_buffer.commit()

        else:
            await connection.conversation.item.create(
                item={
                    "type": "message",
                    "role": "user",
                    "status": "completed" if chunk.eod else "incomplete",
                    "content": self._item_content_parts(chunk.content),
                },
            )

    async def _send_context(
        self,
        context: LMMContext,
        /,
        *,
        connection: AsyncRealtimeConnection,
    ) -> None:
        for element in context:
            match element:
                case LMMInput() as input_element:
                    await connection.conversation.item.create(
                        item={
                            "type": "message",
                            "status": "completed",
                            "role": "user",
                            "content": self._item_content_parts(input_element.content),
                        },
                    )

                case LMMCompletion() as completion_element:
                    await connection.conversation.item.create(
                        item={
                            "type": "message",
                            "status": "completed",
                            "role": "assistant",
                            "content": self._item_content_parts(completion_element.content),
                        },
                    )

                case LMMToolRequests() as tool_requests:
                    for request in tool_requests.requests:
                        await connection.conversation.item.create(
                            item={
                                "type": "function_call",
                                "call_id": request.identifier,
                                "arguments": json.dumps(request.arguments),
                            },
                        )

                case LMMToolResponses() as tool_responses:
                    for response in tool_responses.responses:
                        await connection.conversation.item.create(
                            item={
                                "type": "function_call_output",
                                "call_id": response.identifier,
                                "output": self._tool_result(response.content),
                            },
                        )

    def _item_content_parts(
        self,
        content: MultimodalContent,
    ) -> list[ConversationItemContentParam]:
        content_parts: list[ConversationItemContentParam] = []
        for part in content.parts:
            match part:
                case TextContent() as text:
                    content_parts.append(
                        {
                            "type": "input_text",
                            "text": text.text,
                        }
                    )

                case MediaData() as media:
                    match media.media:
                        case "audio":
                            content_parts.append(
                                {
                                    "type": "input_audio",
                                    "audio": b64encode(media.data).decode(),
                                }
                            )

                        case "image":
                            # skip not supported with a log to prevent connection break
                            ctx.log_error(
                                "OpenAI realtime input (image) not supported! Skipping..."
                            )

                        case "video":
                            # skip not supported with a log to prevent connection break
                            ctx.log_error(
                                "OpenAI realtime input (video) not supported! Skipping..."
                            )

                case MediaReference():
                    # skip not supported with a log to prevent connection break
                    ctx.log_error(
                        "OpenAI realtime input (MediaReference) not supported! Skipping..."
                    )

                case MetaContent():
                    # skip not supported with a log to prevent connection break
                    ctx.log_error("OpenAI realtime input (MetaContent) not supported! Skipping...")

                case other:  # treat other as json text
                    content_parts.append(
                        {
                            "type": "input_text",
                            "text": other.to_json(),
                        }
                    )

        return content_parts

    def _tool_result(
        self,
        content: MultimodalContent,
    ) -> str:
        response_output: str = ""
        for part in content.parts:
            match part:
                case TextContent() as text:
                    response_output += text.text

                case MediaData() | MediaReference():
                    # skip not supported with a log to prevent connection break
                    ctx.log_error(
                        "OpenAI realtime function result (media) not supported! Skipping..."
                    )

                case MetaContent():
                    # skip not supported with a log to prevent connection break
                    ctx.log_error(
                        "OpenAI realtime function result (MetaContent) not supported! Skipping..."
                    )

                case other:  # treat other as json text
                    response_output += other.to_json()

        return response_output

    async def _send_tool_response(
        self,
        response: LMMToolResponse,
        /,
        *,
        connection: AsyncRealtimeConnection,
    ) -> None:
        await connection.conversation.item.create(
            item={
                "type": "function_call_output",
                "call_id": response.identifier,
                "output": self._tool_result(response.content),
            },
        )

    async def _handle_input(
        self,
        connection: AsyncRealtimeConnection,
        *,
        input_stream: AsyncIterator[LMMStreamInput],
        output_stream: AsyncQueue[LMMSessionOutput],
    ) -> None:
        try:
            async for element in input_stream:
                match element:
                    case LMMStreamChunk() as chunk:
                        await self._send_input_chunk(
                            chunk,
                            connection=connection,
                        )

                    case LMMToolResponse() as tool_response:
                        await self._send_tool_response(
                            tool_response,
                            connection=connection,
                        )

        except BaseException as exc:
            output_stream.finish(exception=exc)

        else:
            output_stream.finish()



from collections.abc import Callable, Coroutine
from typing import Protocol, final, overload

from draive.commons import Meta, MetaValues
from draive.instructions.types import (
    Instruction,
    InstructionDeclaration,
    InstructionDeclarationArgument,
    InstructionException,
)
from draive.parameters import ParametrizedFunction

__all__ = (
    "InstructionTemplate",
    "instruction",
)


@final
class InstructionTemplate[**Args](ParametrizedFunction[Args, Coroutine[None, None, str]]):
    __slots__ = ("declaration",)

    def __init__(
        self,
        /,
        name: str,
        *,
        description: str | None,
        meta: Meta,
        function: Callable[Args, Coroutine[None, None, str]],
    ) -> None:
        super().__init__(function)

        self.declaration: InstructionDeclaration
        object.__setattr__(
            self,
            "declaration",
            InstructionDeclaration(
                name=name,
                description=description,
                arguments=[  # should we verify if all are strings?
                    InstructionDeclarationArgument(
                        name=parameter.alias or parameter.name,
                        description=parameter.description,
                        required=parameter.required,
                    )
                    for parameter in self._parameters.values()
                ],
                meta=meta,
            ),
        )

    async def resolve(
        self,
        *args: Args.args,
        **kwargs: Args.kwargs,
    ) -> Instruction:
        try:
            return Instruction(
                name=self.declaration.name,
                description=self.declaration.description,
                content=await super().__call__(*args, **kwargs),  # pyright: ignore[reportCallIssue]
                arguments={},
                meta=self.declaration.meta,
            )

        except Exception as exc:
            raise InstructionException(
                f"Resolving instruction '{self.declaration.name}' failed"
            ) from exc


class InstructionTemplateWrapper(Protocol):
    def __call__[**Args](
        self,
        function: Callable[Args, Coroutine[None, None, str]],
    ) -> InstructionTemplate[Args]: ...


@overload
def instruction[**Args](
    function: Callable[Args, Coroutine[None, None, str]],
    /,
) -> InstructionTemplate[Args]: ...


@overload
def instruction(
    *,
    name: str | None = None,
    description: str | None = None,
    meta: Meta | MetaValues | None = None,
) -> InstructionTemplateWrapper: ...


def instruction[**Args](
    function: Callable[Args, Coroutine[None, None, str]] | None = None,
    *,
    name: str | None = None,
    description: str | None = None,
    meta: Meta | MetaValues | None = None,
) -> InstructionTemplateWrapper | InstructionTemplate[Args]:
    def wrap[**Arg](
        function: Callable[Arg, Coroutine[None, None, str]],
    ) -> InstructionTemplate[Arg]:
        return InstructionTemplate[Arg](
            name=name or function.__name__,
            description=description,
            meta=Meta.of(meta),
            function=function,
        )

    if function := function:
        return wrap(function=function)

    else:
        return wrap



from collections.abc import Mapping, MutableMapping
from typing import Any, Self, cast, final

from haiway import MISSING, DefaultValue, Missing
from haiway.state import AttributeAnnotation

from draive.parameters.specification import ParameterSpecification, parameter_specification
from draive.parameters.types import (
    ParameterConversion,
    ParameterValidation,
    ParameterValidationContext,
    ParameterVerification,
)
from draive.parameters.validation import ParameterValidator

__all__ = ("Parameter",)


@final
class Parameter[Type]:
    @classmethod
    def of(
        cls,
        annotation: AttributeAnnotation,
        /,
        *,
        name: str,
        alias: str | None,
        description: str | Missing,
        default: DefaultValue[Type],
        validator: ParameterValidation[Type] | Missing,
        verifier: ParameterVerification[Type] | Missing,
        converter: ParameterConversion[Type] | Missing,
        specification: ParameterSpecification | Missing,
        required: bool,
    ) -> Self:
        assert validator is MISSING or verifier is MISSING  # nosec: B101
        assert description is MISSING or specification is MISSING  # nosec: B101

        return cls(
            name=name,
            description=None if description is MISSING else cast(str, description),
            alias=alias,
            default=default,
            validator=ParameterValidator.of(
                annotation,
                verifier=None
                if verifier is MISSING
                else cast(ParameterVerification[Type], verifier),
                recursion_guard={},  # TODO: add enclosing type Self?
            )
            if validator is MISSING
            else cast(ParameterValidation[Type], validator),
            converter=converter,
            specification=parameter_specification(
                annotation,
                description=cast(str, description) if description is not MISSING else None,
            )
            if specification is MISSING
            else cast(ParameterSpecification, specification),
            required=required,
        )

    def __init__(
        self,
        *,
        name: str,
        description: str | None,
        alias: str | None,
        default: DefaultValue[Type],
        validator: ParameterValidation[Type],
        converter: ParameterConversion[Type] | Missing,
        specification: ParameterSpecification,
        required: bool,
    ) -> None:
        self.name: str = name
        self.description: str | None = description
        self.alias: str | None = alias
        self.default: DefaultValue[Type] = default
        self.validator: ParameterValidation[Any] = validator
        self.converter: ParameterConversion[Type] | None = (
            None if converter is MISSING else cast(ParameterConversion[Type], converter)
        )
        self.specification: ParameterSpecification = specification
        self.required: bool = required

    def find(
        self,
        mapping: Mapping[str, Any],
        /,
    ) -> Any:
        if self.alias:
            return mapping.get(
                self.alias,
                mapping.get(
                    self.name,
                    MISSING,
                ),
            )

        else:
            return mapping.get(
                self.name,
                MISSING,
            )

    def pick(
        self,
        mapping: MutableMapping[str, Any],
        /,
    ) -> Any:
        if self.alias:
            if self.alias in mapping:
                picked: Any = mapping[self.alias]
                del mapping[self.alias]
                return picked

            elif self.name in mapping:
                picked: Any = mapping[self.name]
                del mapping[self.name]
                return picked

            else:
                return MISSING

        elif self.name in mapping:
            picked: Any = mapping[self.name]
            del mapping[self.name]
            return picked

        else:
            return MISSING

    def validated(
        self,
        value: Any,
        /,
        context: ParameterValidationContext,
    ) -> Any:
        with context.scope(f".{self.name}"):
            try:
                return self.validator(
                    value if value is not MISSING else self.default(),
                    context=context,
                )

            except Exception as exc:
                if value is MISSING:
                    raise ValueError("Missing value") from None

                else:
                    raise exc



from asyncio import CancelledError, Task
from collections.abc import AsyncGenerator, AsyncIterator
from typing import (
    Any,
)

from haiway import AsyncQueue, ctx

from draive.instructions import Instruction
from draive.lmm import (
    LMMSession,
    LMMSessionEvent,
    LMMSessionOutput,
    LMMStreamChunk,
    LMMStreamInput,
    LMMToolRequest,
    LMMToolResponse,
)
from draive.multimodal import MetaContent, MultimodalContent
from draive.realtime.types import RealtimeOutputSelection
from draive.tools import Toolbox

__all__ = ("realtime_process",)


async def realtime_process(  # noqa: C901
    *,
    instruction: Instruction | None,
    input_stream: AsyncIterator[MultimodalContent],
    toolbox: Toolbox,
    output: RealtimeOutputSelection,
    **extra: Any,
) -> AsyncIterator[MultimodalContent]:
    input_queue = AsyncQueue[LMMStreamInput]()

    async def process_input() -> None:
        try:
            async for element in input_stream:
                # we are expecting continuous stream of data without turns,
                # this api will work only with LMMSession implementations
                # which can automatically detect turn ends if needed
                input_queue.enqueue(LMMStreamChunk.of(element))

            input_queue.finish()

        except CancelledError:
            # just finish input on cancel
            input_queue.finish()

        except BaseException as exc:
            input_queue.finish(exception=exc)

    pending_tool_requests: dict[str, Task[None]] = {}

    async def handle_tool_request(
        tool_request: LMMToolRequest,
        /,
    ) -> None:
        try:
            response: LMMToolResponse = await toolbox.respond(tool_request)
            # check if task was not cancelled before passing the result to input
            ctx.check_cancellation()
            # deliver the result directly to input
            input_queue.enqueue(response)

        except CancelledError:
            pass  # just cancelled, ignore

        finally:
            # update pending tools - we have delivered or ignored the result
            del pending_tool_requests[tool_request.identifier]

    input_processing_task: Task[None] = ctx.spawn(process_input)

    session_output: AsyncIterator[LMMSessionOutput]
    try:
        session_output = await LMMSession.prepare(
            instruction=instruction,
            input_stream=input_queue,
            output=output,
            tools=toolbox.available_tools(),
        )

    except BaseException as exc:
        input_processing_task.cancel()
        raise exc

    async def process_output() -> AsyncGenerator[MultimodalContent]:
        try:
            async for chunk in session_output:
                match chunk:
                    case LMMStreamChunk() as chunk:
                        yield chunk.content

                    case LMMSessionEvent() as event:
                        match event.category:
                            case "interrupted":
                                # cancel pending tools on interrupt
                                for task in pending_tool_requests.values():
                                    task.cancel()

                                yield MultimodalContent.of(
                                    MetaContent.of(
                                        "interrupted",
                                        meta=event.meta,
                                    )
                                )

                            case "completed":
                                assert not pending_tool_requests  # nosec: B101
                                yield MultimodalContent.of(
                                    MetaContent.of(
                                        "completed",
                                        meta=event.meta,
                                    )
                                )

                            case other:
                                yield MultimodalContent.of(
                                    MetaContent.of(
                                        other,
                                        meta=event.meta,
                                    )
                                )

                    case LMMToolRequest() as tool_request:
                        pending_tool_requests[tool_request.identifier] = ctx.spawn(
                            handle_tool_request,
                            tool_request,
                        )

        finally:
            input_processing_task.cancel()

    return ctx.stream(process_output)



from collections.abc import Iterable
from typing import Any

from draive.generation import ModelGeneration
from draive.parameters import DataModel

__all__ = ("generate_case_parameters",)

INSTRUCTION: str = """\
You are preparing automated evaluation test scenarios.

<INSTRUCTION>
Familiarize yourself with the SCHEMA describing desired data structure.
Prepare diversified and high quality test scenarios according to the provided SCHEMA.
Results should cover various possible valid outcomes to cover multiple possible scenarios.
</INSTRUCTION>
{guidelines}
<SCHEMA>
{{schema}}
</SCHEMA>

<FORMAT>
Provide the result using a single raw valid JSON object that adheres strictly to the given \
SCHEMA without any comments, formatting, or additional elements.
</FORMAT>
"""

INPUT: str = "Prepare additional scenario"


async def generate_case_parameters[Parameters: DataModel](
    parameters: type[Parameters],
    /,
    *,
    count: int,
    examples: Iterable[Parameters],
    guidelines: str | None = None,
) -> list[Parameters]:
    results: list[Parameters] = []
    example_pairs: list[tuple[str, Any]] = [(INPUT, example) for example in examples]

    for _ in range(0, count):
        results.append(
            await ModelGeneration.generate(
                parameters,
                instruction=INSTRUCTION.format(
                    guidelines=f"\n<GUIDELINES>\n{guidelines}\n</GUIDELINES>\n"
                    if guidelines is not None
                    else ""
                ),
                input=INPUT,
                examples=example_pairs,
                schema_injection="full",
            )
        )
        # put each generated element into the examples for next generation
        # to prevent repetitions and encourage diversified results
        example_pairs.append((INPUT, results[-1]))

    return results



from draive.commons.metadata import (
    META_EMPTY,
    Meta,
    MetaTags,
    MetaValue,
    MetaValues,
    validated_meta_value,
)

__all__ = (
    "META_EMPTY",
    "Meta",
    "MetaTags",
    "MetaValue",
    "MetaValues",
    "validated_meta_value",
)



from collections.abc import Callable, Iterable, Mapping
from inspect import Parameter as InspectParameter
from inspect import _empty as INSPECT_EMPTY  # pyright: ignore[reportPrivateUsage]
from inspect import signature
from types import EllipsisType
from typing import Any, ClassVar, cast, final, get_type_hints, overload

from haiway import MISSING, DefaultValue, Missing
from haiway.state import AttributeAnnotation
from haiway.state.attributes import resolve_attribute_annotation
from haiway.utils import mimic_function

from draive.parameters.parameter import Parameter
from draive.parameters.specification import ParameterSpecification
from draive.parameters.types import ParameterValidation, ParameterVerification
from draive.parameters.validation import ParameterValidationContext

__all__ = (
    "Argument",
    "ParametrizedFunction",
)


@overload
def Argument[Value](
    *,
    aliased: str | None = None,
    description: str | Missing = MISSING,
    default: Value | Missing = MISSING,
    validator: ParameterValidation[Value] | Missing = MISSING,
    specification: ParameterSpecification | Missing = MISSING,
) -> Value: ...


@overload
def Argument[Value](
    *,
    aliased: str | None = None,
    description: str | Missing = MISSING,
    default_factory: Callable[[], Value] | Missing = MISSING,
    validator: ParameterValidation[Value] | Missing = MISSING,
    specification: ParameterSpecification | Missing = MISSING,
) -> Value: ...


@overload
def Argument[Value](
    *,
    aliased: str | None = None,
    description: str | Missing = MISSING,
    default: Value | Missing = MISSING,
    verifier: ParameterVerification[Value] | Missing = MISSING,
    specification: ParameterSpecification | Missing = MISSING,
) -> Value: ...


@overload
def Argument[Value](
    *,
    aliased: str | None = None,
    description: str | Missing = MISSING,
    default_factory: Callable[[], Value] | Missing = MISSING,
    verifier: ParameterVerification[Value] | Missing = MISSING,
    specification: ParameterSpecification | Missing = MISSING,
) -> Value: ...


def Argument[Value](
    *,
    aliased: str | None = None,
    description: str | Missing = MISSING,
    default: Value | Missing = MISSING,
    default_factory: Callable[[], Value] | Missing = MISSING,
    validator: ParameterValidation[Value] | Missing = MISSING,
    verifier: ParameterVerification[Value] | Missing = MISSING,
    specification: ParameterSpecification | Missing = MISSING,
) -> Value:  # it is actually a FunctionArgument, but type checker has to be fooled
    assert (  # nosec: B101
        default is MISSING or default_factory is MISSING
    ), "Can't specify both default value and factory"
    assert (  # nosec: B101
        description is MISSING or specification is MISSING
    ), "Can't specify both description and specification"
    assert (  # nosec: B101
        validator is MISSING or verifier is MISSING
    ), "Can't specify both validator and verifier"

    return cast(
        Value,
        FunctionArgument(
            aliased=aliased,
            description=description,
            default=default,
            default_factory=default_factory,
            validator=validator,
            verifier=verifier,
            specification=specification,
        ),
    )


@final
class FunctionArgument:
    def __init__(
        self,
        aliased: str | None,
        description: str | Missing,
        default: Any | Missing,
        default_factory: Callable[[], Any] | Missing,
        validator: ParameterValidation[Any] | Missing,
        verifier: ParameterVerification[Any] | Missing,
        specification: ParameterSpecification | Missing,
    ) -> None:
        self.aliased: str | None = aliased
        self.description: str | Missing = description
        self.default: Any | Missing = default
        self.default_factory: Callable[[], Any] | Missing = default_factory
        self.validator: ParameterValidation[Any] | Missing = validator
        self.verifier: ParameterVerification[Any] | Missing = verifier
        self.specification: ParameterSpecification | Missing = specification


class ParametrizedFunction[**Args, Result]:
    __IMMUTABLE__: ClassVar[EllipsisType] = ...

    __slots__ = (
        "__defaults__",
        "__doc__",
        "__globals__",
        "__kwdefaults__",
        "__name__",
        "__qualname__",
        "__wrapped__",
        "_call",
        "_name",
        "_parameters",
        "_variadic_keyword_parameters",
        "_write",
    )

    def __init__(
        self,
        function: Callable[Args, Result],
        /,
    ) -> None:
        assert (  # nosec: B101
            not isinstance(function, ParametrizedFunction)
        ), "Cannot parametrize the same function more than once!"

        self._call: Callable[Args, Result]
        object.__setattr__(
            self,
            "_call",
            function,
        )
        self._name: str
        object.__setattr__(
            self,
            "_name",
            function.__name__,
        )
        self._parameters: dict[str, Parameter[Any]]
        object.__setattr__(
            self,
            "_parameters",
            {},
        )
        self._variadic_keyword_parameters: Parameter[Any] | None
        object.__setattr__(
            self,
            "_variadic_keyword_parameters",
            None,
        )
        type_hints: Mapping[str, Any] = get_type_hints(function)
        for parameter in signature(function).parameters.values():
            match parameter.kind:
                case InspectParameter.POSITIONAL_ONLY | InspectParameter.VAR_POSITIONAL:
                    raise NotImplementedError("Positional only arguments are not supported yet.")

                case InspectParameter.VAR_KEYWORD:
                    assert self._variadic_keyword_parameters is None  # nosec: B101
                    object.__setattr__(
                        self,
                        "_variadic_keyword_parameters",
                        _resolve_argument(
                            parameter,
                            module=function.__module__,
                            type_hint=type_hints.get(parameter.name),
                        ),
                    )

                case _:
                    self._parameters[parameter.name] = _resolve_argument(
                        parameter,
                        module=function.__module__,
                        type_hint=type_hints.get(parameter.name),
                    )

        mimic_function(function, within=self)

    @property
    def arguments(self) -> Iterable[Parameter[Any]]:
        return self._parameters.values()

    def validate_arguments(
        self,
        **kwargs: Any,
    ) -> dict[str, Any]:
        # TODO: add support for positional arguments
        with ParameterValidationContext().scope(self.__class__.__qualname__) as context:
            validated: dict[str, Any] = {}
            if self._variadic_keyword_parameters is None:
                for parameter in self._parameters.values():
                    validated[parameter.name] = parameter.validated(
                        parameter.find(kwargs),
                        context=context,
                    )

            else:
                for parameter in self._parameters.values():
                    validated[parameter.name] = parameter.validated(
                        parameter.pick(kwargs),
                        context=context,
                    )

                for key, value in kwargs.items():
                    validated[key] = self._variadic_keyword_parameters.validated(
                        value,
                        context=context,
                    )

            return validated

    def __call__(
        self,
        *args: Args.args,
        **kwargs: Args.kwargs,
    ) -> Result:
        assert not args, "Positional unkeyed arguments are not supported"  # nosec: B101
        return self._call(*args, **self.validate_arguments(**kwargs))  # pyright: ignore[reportCallIssue]

    def __setattr__(
        self,
        name: str,
        value: Any,
    ) -> Any:
        raise AttributeError(
            f"Can't modify immutable {self.__class__.__qualname__},"
            f" attribute - '{name}' cannot be modified"
        )

    def __delattr__(
        self,
        name: str,
    ) -> None:
        raise AttributeError(
            f"Can't modify immutable {self.__class__.__qualname__},"
            f" attribute - '{name}' cannot be deleted"
        )


def _resolve_argument(
    parameter: InspectParameter,
    /,
    *,
    module: str,
    type_hint: Any,
) -> Parameter[Any]:
    if parameter.annotation is INSPECT_EMPTY or type_hint is None:
        raise TypeError(
            "Untyped argument %s",
            parameter.name,
        )

    attribute: AttributeAnnotation = resolve_attribute_annotation(
        type_hint,
        module=module,
        type_parameters={},
        self_annotation=None,
        recursion_guard={},
    )

    match parameter.default:
        case FunctionArgument() as argument:
            return Parameter[Any].of(
                attribute,
                name=parameter.name,
                alias=argument.aliased,
                description=argument.description,
                default=DefaultValue(
                    argument.default,
                    factory=argument.default_factory,
                ),
                validator=argument.validator,
                verifier=argument.verifier,
                converter=MISSING,
                specification=argument.specification,
                required=attribute.required
                and argument.default is MISSING
                and argument.default_factory is MISSING,
            )

        case DefaultValue() as default:  # pyright: ignore[reportUnknownVariableType]
            return Parameter[Any].of(
                attribute,
                name=parameter.name,
                alias=None,
                description=MISSING,
                default=default,  # pyright: ignore[reportUnknownArgumentType]
                validator=MISSING,
                verifier=MISSING,
                converter=MISSING,
                specification=MISSING,
                required=False,
            )

        case value:
            return Parameter[Any].of(
                attribute,
                name=parameter.name,
                alias=None,
                description=MISSING,
                default=DefaultValue(MISSING if value is INSPECT_EMPTY else value),
                validator=MISSING,
                verifier=MISSING,
                converter=MISSING,
                specification=MISSING,
                required=attribute.required and value is INSPECT_EMPTY,
            )



from pathlib import Path

from draive.gemini.api import GeminiAPI
from draive.sentencepiece import sentencepiece_processor, sentencepiece_tokenizer
from draive.tokenization import Tokenization

__all__ = ("GeminiTokenization",)


class GeminiTokenization(GeminiAPI):
    async def tokenizer(
        self,
        model_name: str,
        /,
    ) -> Tokenization:
        """
        Prepare tokenizer for selected Gemini model.
        """

        return Tokenization(
            text_tokenizing=sentencepiece_tokenizer(
                await sentencepiece_processor(model_path=_model_path(model_name))
            )
        )


def _model_path(model_name: str) -> str:
    model_file: str = _mapping.get(model_name, "gemini_tokenizer.model")
    return str(Path(__file__).parent / "tokens" / model_file)


_mapping: dict[str, str] = {}  # empty, using only default tokenizer



from typing import cast

from haiway import MISSING, Missing

__all__ = ("unwrap_missing",)


def unwrap_missing[Value, Default](
    value: Value | Missing,
    /,
) -> Value | None:
    if value is MISSING:
        return None
    else:
        return cast(Value, value)



from collections.abc import Sequence
from typing import Any

from haiway import State, ctx

from draive.tokenization.types import TextTokenizing

__all__ = ("Tokenization",)


class Tokenization(State):
    @classmethod
    def tokenize(
        cls,
        text: str,
        **extra: Any,
    ) -> Sequence[int]:
        """
        Tokenize input text using current Tokenization state.
        """

        return ctx.state(cls).text_tokenizing(
            text,
            **extra,
        )

    @classmethod
    def count_tokens(
        cls,
        text: str,
        **extra: Any,
    ) -> int:
        """
        Count input text tokens using current Tokenization state.
        """

        return len(
            ctx.state(cls).text_tokenizing(
                text,
                **extra,
            )
        )

    text_tokenizing: TextTokenizing



from collections.abc import AsyncGenerator, AsyncIterator, Iterable

from haiway import State, ctx

from draive.instructions import Instruction
from draive.multimodal import Multimodal, MultimodalContent
from draive.realtime.default import realtime_process
from draive.realtime.types import RealtimeOutputSelection, RealtimeProcessing
from draive.tools import Tool, Toolbox

__all__ = ("Realtime",)


class Realtime(State):
    @classmethod
    async def process(
        cls,
        *,
        instruction: Instruction | str | None = None,
        input: AsyncIterator[Multimodal],  # noqa: A002
        tools: Toolbox | Iterable[Tool] | None = None,
        output: RealtimeOutputSelection = "auto",
    ) -> AsyncIterator[MultimodalContent]:
        async def input_stream() -> AsyncGenerator[MultimodalContent]:
            async for element in input:
                yield MultimodalContent.of(element)

        return await ctx.state(cls).processing(
            instruction=Instruction.of(instruction),
            input_stream=input_stream(),
            toolbox=Toolbox.of(tools),
            output=output,
        )

    processing: RealtimeProcessing = realtime_process



from collections.abc import Sequence
from typing import Any, Protocol, runtime_checkable

from draive.commons import Meta, MetaValues
from draive.multimodal import MultimodalContent

__all__ = (
    "GuardrailsInputModerationException",
    "GuardrailsModerationChecking",
    "GuardrailsModerationException",
    "GuardrailsOutputModerationException",
)


class GuardrailsModerationException(Exception):
    def __init__(
        self,
        *args: object,
        violations: Sequence[str],
        content: MultimodalContent,
        replacement: MultimodalContent | None = None,
        meta: Meta | MetaValues | None = None,
    ) -> None:
        super().__init__(*args)
        self.violations: Sequence[str] = violations
        self.content: MultimodalContent = content
        self.replacement: MultimodalContent | None = replacement
        self.meta: Meta = Meta.of(meta)


class GuardrailsInputModerationException(GuardrailsModerationException):
    def __init__(
        self,
        *args: object,
        violations: Sequence[str],
        content: MultimodalContent,
        replacement: MultimodalContent | None = None,
        meta: Meta | MetaValues | None = None,
    ) -> None:
        super().__init__(
            *args,
            violations=violations,
            content=content,
            replacement=replacement,
            meta=meta,
        )


class GuardrailsOutputModerationException(GuardrailsModerationException):
    def __init__(
        self,
        *args: object,
        violations: Sequence[str],
        content: MultimodalContent,
        replacement: MultimodalContent | None = None,
        meta: Meta | MetaValues | None = None,
    ) -> None:
        super().__init__(
            *args,
            violations=violations,
            content=content,
            replacement=replacement,
            meta=meta,
        )


@runtime_checkable
class GuardrailsModerationChecking(Protocol):
    async def __call__(
        self,
        content: MultimodalContent,
        /,
        **extra: Any,
    ) -> None: ...



from draive.anthropic.client import Anthropic
from draive.anthropic.config import AnthropicConfig
from draive.anthropic.types import AnthropicException

__all__ = (
    "Anthropic",
    "AnthropicConfig",
    "AnthropicException",
)



import json
from collections.abc import AsyncIterator, Iterable
from itertools import chain
from typing import Any, Literal, overload
from uuid import uuid4

from haiway import ObservabilityLevel, ctx
from ollama import ChatResponse, Message, Options

from draive.instructions import Instruction
from draive.lmm import (
    LMM,
    LMMCompletion,
    LMMContext,
    LMMOutput,
    LMMOutputSelection,
    LMMToolRequest,
    LMMToolRequests,
    LMMToolSelection,
    LMMToolSpecification,
)
from draive.lmm.types import LMMStreamOutput
from draive.multimodal import MultimodalContent
from draive.ollama.api import OllamaAPI
from draive.ollama.config import OllamaChatConfig
from draive.ollama.lmm import (
    context_element_as_messages,
    output_as_response_declaration,
    tools_as_tool_config,
)
from draive.ollama.types import OllamaException
from draive.ollama.utils import unwrap_missing

__all__ = ("OllamaLMMGeneration",)


class OllamaLMMGeneration(OllamaAPI):
    def lmm(self) -> LMM:
        return LMM(completing=self.lmm_completion)

    @overload
    async def lmm_completion(
        self,
        *,
        instruction: Instruction | None,
        context: LMMContext,
        tool_selection: LMMToolSelection,
        tools: Iterable[LMMToolSpecification] | None,
        config: OllamaChatConfig | None = None,
        output: LMMOutputSelection,
        stream: Literal[False] = False,
        **extra: Any,
    ) -> LMMOutput: ...

    @overload
    async def lmm_completion(
        self,
        *,
        instruction: Instruction | None,
        context: LMMContext,
        tool_selection: LMMToolSelection,
        tools: Iterable[LMMToolSpecification] | None,
        output: LMMOutputSelection,
        config: OllamaChatConfig | None = None,
        stream: Literal[True],
        **extra: Any,
    ) -> AsyncIterator[LMMStreamOutput]: ...

    async def lmm_completion(
        self,
        *,
        instruction: Instruction | None,
        context: LMMContext,
        tool_selection: LMMToolSelection,
        tools: Iterable[LMMToolSpecification] | None,
        output: LMMOutputSelection,
        config: OllamaChatConfig | None = None,
        stream: bool = False,
        **extra: Any,
    ) -> AsyncIterator[LMMStreamOutput] | LMMOutput:
        if stream:
            raise NotImplementedError("ollama streaming is not implemented yet")

        chat_config: OllamaChatConfig = config or ctx.state(OllamaChatConfig)
        with ctx.scope("ollama_lmm_completion", chat_config):
            ctx.record(
                ObservabilityLevel.INFO,
                attributes={
                    "lmm.provider": "ollama",
                    "lmm.model": chat_config.model,
                    "lmm.temperature": chat_config.temperature,
                    "lmm.max_tokens": chat_config.max_tokens,
                    "lmm.seed": chat_config.seed,
                    "lmm.tools": [tool["name"] for tool in tools] if tools else None,
                    "lmm.tool_selection": f"{tool_selection}" if tools else None,
                    "lmm.stream": stream,
                    "lmm.output": f"{output}",
                    "lmm.instruction": f"{instruction}",
                    "lmm.context": [element.to_str() for element in context],
                },
            )

            messages: list[Message] = list(
                chain.from_iterable([context_element_as_messages(element) for element in context])
            )

            if instruction:
                messages = [
                    Message(
                        role="system",
                        content=Instruction.formatted(instruction),
                    ),
                    *messages,
                ]

            response_format, output_decoder = output_as_response_declaration(output)

            tools_list = tools_as_tool_config(
                tools,
                tool_selection=tool_selection,
            )

            completion: ChatResponse = await self._client.chat(
                model=chat_config.model,
                messages=messages,
                format=response_format,
                tools=tools_list,
                options=Options(
                    temperature=chat_config.temperature,
                    num_predict=unwrap_missing(chat_config.max_tokens),
                    top_k=unwrap_missing(chat_config.top_k),
                    top_p=unwrap_missing(chat_config.top_p),
                    seed=unwrap_missing(chat_config.seed),
                    stop=unwrap_missing(chat_config.stop_sequences),
                ),
                stream=False,
            )

            completion_message: Message = completion.message

            lmm_completion: LMMCompletion | None
            if content := completion_message.content:
                lmm_completion = LMMCompletion.of(output_decoder(MultimodalContent.of(content)))

            else:
                lmm_completion = None

            if tool_calls := completion_message.tool_calls:
                assert tools, "Requesting tool call without tools"  # nosec: B101
                completion_tool_calls = LMMToolRequests(
                    content=lmm_completion.content if lmm_completion else None,
                    requests=[
                        LMMToolRequest(
                            identifier=uuid4().hex,
                            tool=call.function.name,
                            arguments=json.loads(call.function.arguments)
                            if isinstance(call.function.arguments, str)
                            else call.function.arguments,
                        )
                        for call in tool_calls
                    ],
                )
                ctx.record(
                    ObservabilityLevel.INFO,
                    event="lmm.tool_requests",
                    attributes={"lmm.tools": [call.function.name for call in tool_calls]},
                )
                return completion_tool_calls

            elif lmm_completion:
                ctx.record(
                    ObservabilityLevel.INFO,
                    event="lmm.completion",
                )
                return lmm_completion

            else:
                raise OllamaException("Invalid Ollama completion, missing content!", completion)



from collections.abc import Sequence
from typing import Literal

from haiway import MISSING, Missing

from draive.configuration import Config

__all__ = (
    "GeminiEmbeddingConfig",
    "GeminiGenerationConfig",
    "GeminiLiveConfig",
)


class GeminiGenerationConfig(Config):
    model: str
    temperature: float = 1.0
    top_p: float | Missing = MISSING
    top_k: int | Missing = MISSING
    max_tokens: int | Missing = MISSING
    seed: int | Missing = MISSING
    stop_sequences: Sequence[str] | Missing = MISSING
    speech_voice_name: str | Missing = MISSING
    media_resolution: Literal["low", "medium", "high"] | Missing = MISSING


class GeminiLiveConfig(Config):
    model: str
    temperature: float = 1.0
    top_p: float | Missing = MISSING
    top_k: int | Missing = MISSING
    max_tokens: int | Missing = MISSING
    seed: int | Missing = MISSING
    stop_sequences: Sequence[str] | Missing = MISSING
    speech_voice_name: str | Missing = MISSING
    media_resolution: Literal["low", "medium", "high"] | Missing = MISSING


class GeminiEmbeddingConfig(Config):
    model: str
    dimensions: int | Missing = MISSING
    batch_size: int = 128



import pytest

from draive.splitters.exhaustive import exhaustive_regex_split_text


@pytest.fixture
def sample_text() -> str:
    with open("tests/data/markdown_sample_text.txt") as file:
        return file.read()


@pytest.fixture
def exhaustive_splitting_expected_result() -> str:
    with open("tests/data/markdown_exhaustive_splitting.txt") as file:
        return file.read()


@pytest.fixture
def exhaustive_overlap_splitting_expected_result() -> str:
    with open("tests/data/markdown_exhaustive_overlap_splitting.txt") as file:
        return file.read()


def test_returns_single_part_when_text_fits_part_size() -> None:
    assert exhaustive_regex_split_text(
        text="0",
        part_size=1,
        count_size=len,
        separators=[],
    ) == ["0"]


def test_fails_when_no_separators_provided() -> None:
    with pytest.raises(ValueError):
        exhaustive_regex_split_text(
            text="0",
            part_size=0,
            count_size=len,
            separators=[],
        )


def test_fails_when_cant_split() -> None:
    with pytest.raises(ValueError):
        exhaustive_regex_split_text(
            text="01",
            part_size=1,
            count_size=len,
            separators=[r"\s"],
        )


def test_strips_whitespaces_with_leading_and_trailing_part_whitespaces() -> None:
    assert exhaustive_regex_split_text(
        text=" 01  23 ",
        part_size=3,
        count_size=len,
        separators=[r"\s"],
    ) == ["01", "23"]


def test_adds_leading_splitters_when_required() -> None:
    assert exhaustive_regex_split_text(
        text="#01#23",
        part_size=3,
        count_size=len,
        separators=["#"],
    ) == ["#01", "#23"]


def test_uses_paragraphs_splitting_with_default_separators() -> None:
    assert exhaustive_regex_split_text(
        text="01\n\n23",
        part_size=4,
        count_size=len,
    ) == ["01", "23"]


def test_uses_newline_splitting_with_default_separators() -> None:
    assert exhaustive_regex_split_text(
        text="01\n23",
        part_size=4,
        count_size=len,
    ) == ["01", "23"]


def test_uses_words_splitting_with_default_separators() -> None:
    assert exhaustive_regex_split_text(
        text="01 23",
        part_size=3,
        count_size=len,
    ) == ["01", "23"]


def test_uses_words_splitting_when_one_separators_provided() -> None:
    assert exhaustive_regex_split_text(
        text="01 23",
        part_size=3,
        count_size=len,
        separators="#",
    ) == ["01", "23"]


def test_splitting_text_when_parts_fit_part_size_exactly() -> None:
    assert exhaustive_regex_split_text(
        text="#ab#cd#ef#gh",
        part_size=3,
        count_size=len,
        separators="#",
    ) == ["#ab", "#cd", "#ef", "#gh"]


def test_splitting_markdown_headers() -> None:
    assert exhaustive_regex_split_text(
        text="# Header1\n\nThis is Header1\n\n# Header2\n\nThis is Header2",
        part_size=30,
        count_size=len,
        separators=[r"#\s.*"],
    ) == [
        "# Header1\n\nThis is Header1",
        "# Header2\n\nThis is Header2",
    ]


def test_splitting_markdown_headers_with_subheaders() -> None:
    assert exhaustive_regex_split_text(
        text="# Title\n\nThis is Header1\n\n# Subtitle\n\nThis is Sub-header",
        part_size=30,
        count_size=len,
        separators=[r"#{1,6}\s.*"],
    ) == [
        "# Title\n\nThis is Header1",
        "# Subtitle\n\nThis is Sub-header",
    ]


def test_splitting_html_tag() -> None:
    assert exhaustive_regex_split_text(
        text='<div class="container">This is a short text</div>',
        part_size=45,
        count_size=len,
        separators=[r"<.+?>"],
    ) == [
        '<div class="container">This is a short text',
        "</div>",
    ]


def test_overlapping_parts_with_parts_similar_sizes() -> None:
    assert exhaustive_regex_split_text(
        text="#a#b#c#d#e#f#g#h#i#j#k#l#m",
        part_size=9,
        count_size=len,
        separators=["#"],
        part_overlap_size=6,
    ) == ["#a#b#c#d", "#c#d#e#f", "#e#f#g#h", "#g#h#i#j", "#i#j#k#l", "#k#l#m"]


def test_overlapping_parts_with_different_sizes() -> None:
    assert exhaustive_regex_split_text(
        text="#1234 56#788#90#ab#cd#ef#g#h ijklm",
        part_size=6,
        count_size=len,
        separators=[r"\s", "#"],
        part_overlap_size=6,
    ) == [
        "#1234",
        "#1234 56",
        "56#788",
        "#788#90",
        "#90#ab",
        "#ab#cd",
        "#cd#ef",
        "#ef#g",
        "#ef#g#h",
        "#h ijklm",
    ]


def test_regex_text_splitting_with_long_text(
    sample_text: str,
    exhaustive_splitting_expected_result: str,
) -> None:
    result: list[str] = exhaustive_regex_split_text(
        text=sample_text,
        part_size=500,
        count_size=len,
        separators=r"#{1,6}\s.*",
    )
    assert "\n\n\n\n".join(result) == exhaustive_splitting_expected_result


def test_uses_overlap_splitting_with_long_text(
    sample_text: str,
    exhaustive_overlap_splitting_expected_result: str,
) -> None:
    result: list[str] = exhaustive_regex_split_text(
        text=sample_text,
        part_size=500,
        count_size=len,
        separators=r"#{1,6}\s.*",
        part_overlap_size=100,
    )
    assert "\n\n\n\n".join(result) == exhaustive_overlap_splitting_expected_result



from collections.abc import Collection, Iterable, Mapping
from types import TracebackType
from typing import Any, Literal, final

from haiway import State

from draive.vllm.api import VLLMAPI
from draive.vllm.embedding import VLLMEmbedding
from draive.vllm.lmm_generation import VLLMLMMGeneration

__all__ = ("VLLM",)


@final
class VLLM(
    VLLMLMMGeneration,
    VLLMEmbedding,
    VLLMAPI,
):
    """
    Access to VLLM services, can be used to prepare various functionalities like lmm.
    """

    __slots__ = ("_features",)

    def __init__(
        self,
        base_url: str | None = None,
        default_headers: Mapping[str, str] | None = None,
        features: Collection[Literal["lmm", "text_embedding"]] | None = None,
        **extra: Any,
    ) -> None:
        super().__init__(
            base_url=base_url,
            default_headers=default_headers,
            **extra,
        )
        self._features: frozenset[Literal["lmm", "text_embedding"]] = (
            frozenset(features) if features is not None else frozenset(("lmm", "text_embedding"))
        )

    async def __aenter__(self) -> Iterable[State]:
        await self._initialize_client()
        state: list[State] = []
        if "lmm" in self._features:
            state.append(self.lmm())

        if "text_embedding" in self._features:
            state.append(self.text_embedding())

        return state

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        await self._deinitialize_client()



from draive.parameters.function import Argument, ParametrizedFunction
from draive.parameters.model import DataModel, Field
from draive.parameters.parameter import Parameter
from draive.parameters.specification import (
    ParameterSpecification,
    ParametersSpecification,
    validated_specification,
)
from draive.parameters.types import BasicValue, ParameterValidationContext, ParameterValidationError
from draive.parameters.validation import ParameterValidator, ParameterVerification

__all__ = (
    "Argument",
    "BasicValue",
    "DataModel",
    "Field",
    "Parameter",
    "ParameterSpecification",
    "ParameterValidationContext",
    "ParameterValidationError",
    "ParameterValidator",
    "ParameterVerification",
    "ParametersSpecification",
    "ParametrizedFunction",
    "validated_specification",
)



from collections.abc import Mapping
from typing import Any

from haiway import getenv_str
from openai import AsyncOpenAI

__all__ = ("VLLMAPI",)


class VLLMAPI:
    __slots__ = (
        "_base_url",
        "_client",
        "_default_headers",
        "_extra",
    )

    def __init__(
        self,
        base_url: str | None = None,
        default_headers: Mapping[str, str] | None = None,
        **extra: Any,
    ) -> None:
        self._base_url: str | None = base_url or getenv_str("VLLM_BASE_URL")
        self._default_headers: Mapping[str, str] | None = default_headers
        self._extra: Mapping[str, Any] = extra
        self._client: AsyncOpenAI = self._prepare_client()

    def _prepare_client(self) -> AsyncOpenAI:
        return AsyncOpenAI(
            # using placeholders to prevent reading environment variables targeting actual OpenAI
            organization="vllm",
            project="vllm",
            base_url=self._base_url,
            default_headers=self._default_headers,
            **self._extra,
        )

    async def _initialize_client(self) -> None:
        await self._client.close()
        self._client = self._prepare_client()
        await self._client.__aenter__()

    async def _deinitialize_client(self) -> None:
        await self._client.__aexit__(
            None,
            None,
            None,
        )



from collections.abc import Sequence
from typing import Any, Protocol, runtime_checkable

__all__ = ("TextTokenizing",)


@runtime_checkable
class TextTokenizing(Protocol):
    def __call__(
        self,
        text: str,
        /,
        **extra: Any,
    ) -> Sequence[int]: ...



from collections.abc import Sequence
from typing import Literal, TypedDict

from haiway import MISSING, Missing

from draive.configuration import Config

__all__ = (
    "OpenAIChatConfig",
    "OpenAIEmbeddingConfig",
    "OpenAIImageGenerationConfig",
    "OpenAIModerationConfig",
    "OpenAIRealtimeConfig",
)


class AudioResponseFormat(TypedDict):
    format: Literal["wav", "mp3"]
    voice: Literal["ash", "ballad", "coral", "sage", "verse"]


class OpenAIChatConfig(Config):
    model: str
    temperature: float = 1.0
    top_p: float | Missing = MISSING
    frequency_penalty: float | Missing = MISSING
    max_tokens: int | Missing = MISSING
    seed: int | Missing = MISSING
    audio_response_format: AudioResponseFormat | Missing = MISSING
    vision_details: Literal["auto", "low", "high"] | Missing = MISSING
    parallel_tool_calls: bool | Missing = MISSING
    timeout: float | Missing = MISSING
    stop_sequences: Sequence[str] | Missing = MISSING


class OpenAIRealtimeConfig(Config):
    model: str
    temperature: float = 1.0
    max_tokens: int | Missing = MISSING
    input_audio_format: Literal["pcm16", "g711_ulaw", "g711_alaw"] | Missing = MISSING
    output_audio_format: Literal["pcm16", "g711_ulaw", "g711_alaw"] = "pcm16"
    voice: str | Missing = MISSING
    vad_type: Literal["server_vad", "semantic_vad"] | Missing = MISSING
    vad_eagerness: Literal["low", "medium", "high", "auto"] = "auto"
    transcribe_audio: bool | Missing = MISSING


class OpenAIEmbeddingConfig(Config):
    model: str
    dimensions: int | Missing = MISSING
    batch_size: int = 128
    timeout: float | Missing = MISSING


class OpenAIImageGenerationConfig(Config):
    model: str
    result: Literal["url", "b64_json"]
    quality: Literal["standard", "hd"] = "standard"
    size: Literal["256x256", "512x512", "1024x1024", "1792x1024", "1024x1792"] = "1024x1024"
    style: Literal["vivid", "natural"] = "vivid"
    timeout: float | Missing = MISSING


class OpenAIModerationConfig(Config):
    model: str = "omni-moderation-latest"
    timeout: float | Missing = MISSING
    harassment_threshold: float | Missing = MISSING
    harassment_threatening_threshold: float | Missing = MISSING
    hate_threshold: float | Missing = MISSING
    hate_threatening_threshold: float | Missing = MISSING
    self_harm_threshold: float | Missing = MISSING
    self_harm_instructions_threshold: float | Missing = MISSING
    self_harm_intent_threshold: float | Missing = MISSING
    sexual_threshold: float | Missing = MISSING
    sexual_minors_threshold: float | Missing = MISSING
    violence_threshold: float | Missing = MISSING
    violence_graphic_threshold: float | Missing = MISSING
    illicit_threshold: float | Missing = MISSING
    illicit_violent_threshold: float | Missing = MISSING
